import uuid

from django.db.models.signals import post_delete
from django.dispatch import receiver

from appchance_sections.models import DynamicContentAbstract
from appchance_sections.utils import get_section_model

Section = get_section_model()


@receiver(post_delete, dispatch_uid=uuid.uuid4())
def do_after_delete_dynamic_content(sender, instance, **kwargs):  # pylint: disable=W0613
    if isinstance(instance, DynamicContentAbstract):
        Section.objects.filter(content_slug=instance.content_slug).delete()



import importlib
from typing import Type

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured

from appchance_sections.conf import SECTION_CONTENT_MODEL, SECTION_MODEL
from appchance_sections.models import DynamicContentAbstract


def get_section_model():
    """
    Return the Section model that is active in this project.
    """
    try:
        return django_apps.get_model(SECTION_MODEL, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("SECTION_MODEL must be of the form 'app_label.model_name'") from None
    except LookupError:
        raise ImproperlyConfigured(
            "SECTION_MODEL refers to model '%s' that has not been installed" % SECTION_MODEL
        ) from None
    except AttributeError:
        raise ImproperlyConfigured("SECTION_MODEL must be defined in settings.") from None


Section = get_section_model()


def get_content_model():
    """
    Return the Content model that is active in this project.
    """

    try:
        parts = SECTION_CONTENT_MODEL.split(".")
        module_path, class_name = ".".join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError:
        raise ImproperlyConfigured("SECTION_CONTENT_MODEL must be of the form 'app_label.path.model_name'") from None


Content = get_content_model()


def register_section_content(
    slug, name, url, *, query_params=None, path_params=None, widgets=None, placements=None, prefix=None
):
    SectionRegistry.register_content(
        slug=slug,
        name=name,
        url=url,
        query_params=query_params,
        path_params=path_params,
        widgets=widgets,
        placements=placements,
        prefix=prefix,
    )


def register_dynamic_content(content_class: Type[DynamicContentAbstract]):
    SectionRegistry.register_dynamic_content(content_class=content_class)


def register_content_callback(callback):
    SectionRegistry.register_content_callback(callback)


def unregister_contents():
    SectionRegistry.unregister_contents()


def unregister_dynamic_contents():
    SectionRegistry.unregister_dynamic_contents()


class SectionRegistry:
    _contents = Section._contents  # pylint: disable=W0212
    _content_callbacks = Section._content_callbacks  # pylint: disable=W0212

    @classmethod
    def register_content(
        cls, slug, name, url, *, query_params=None, path_params=None, widgets=None, placements=None, prefix=None
    ):
        content = Content(
            slug=slug,
            name=name,
            url=url,
            query_params=query_params,
            path_params=path_params,
            widgets=widgets,
            placements=placements,
            prefix=prefix,
        )
        cls._contents[content.slug] = content

    @classmethod
    def register_dynamic_content(cls, content_class: Type[DynamicContentAbstract]):
        def _callback():
            contents = content_class.get_all_objects()
            items = []
            for content in contents:
                query_params = content_class.QUERY_PARAMS or {}
                if content_class.FILTER_ATTRIBUTE:
                    query_params[content_class.FILTER_ATTRIBUTE] = content.id
                path_params = None if content_class.FILTER_ATTRIBUTE else {"pk": content.id}

                items.append(
                    Content(
                        slug=content.content_slug,
                        name=content.name_field,
                        url=content_class.URL,
                        query_params=query_params,
                        path_params=path_params,
                        widgets=content_class.WIDGETS,
                        placements=content_class.PLACEMENTS,
                        prefix=(content_class.PREFIX or content_class.__name__),
                    )
                )
            return items

        cls.register_content_callback(callback=_callback)

    @classmethod
    def register_content_callback(cls, callback):
        cls._content_callbacks.add(callback)

    @classmethod
    def unregister_contents(cls):
        cls._contents.clear()

    @classmethod
    def unregister_dynamic_contents(cls):
        cls._content_callbacks.clear()



from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured

from appchance_sections.conf import (
    SECTION_CONTENTS_CACHE_ALIAS,
    SECTION_CONTENTS_CACHE_KEY,
    SECTION_CONTENTS_CACHE_TIMEOUT,
)


def get_contents_cache():
    if SECTION_CONTENTS_CACHE_ALIAS != "default" and SECTION_CONTENTS_CACHE_ALIAS not in settings.CACHES:
        raise ImproperlyConfigured(
            "Cache alias '%s' not set in settings.CACHES" % SECTION_CONTENTS_CACHE_ALIAS
        ) from None
    return caches[SECTION_CONTENTS_CACHE_ALIAS]


def get_contents_cache_key():
    return SECTION_CONTENTS_CACHE_KEY


def get_contents_cache_timeout():
    return int(SECTION_CONTENTS_CACHE_TIMEOUT)



from django.conf import settings

SECTION_MODEL = getattr(settings, "SECTION_MODEL")
SECTION_CONTENT_MODEL = getattr(settings, "SECTION_CONTENT_MODEL", "appchance_sections.models.Content")
SECTION_CONTENTS_CACHE_ALIAS = getattr(settings, "SECTION_CONTENTS_CACHE_ALIAS", "default")
SECTION_CONTENTS_CACHE_KEY = getattr(settings, "SECTION_CONTENTS_CACHE_KEY", "section_content_cache_key")
SECTION_CONTENTS_CACHE_TIMEOUT = getattr(settings, "SECTION_CONTENTS_CACHE_TIMEOUT", 30)



from rest_framework import serializers

from appchance_sections.utils import get_section_model

Section = get_section_model()


class SectionSerializer(serializers.ModelSerializer):
    name = serializers.SerializerMethodField(method_name="get_name")
    content_url = serializers.SerializerMethodField(method_name="get_content_url")

    class Meta:
        model = Section
        fields = ("id", "name", "widget", "placement", "order", "content_url")
        ref_name = "MainSectionSerializer"

    def get_content_url(self, instance):
        request = self.context.get("request")
        kwargs = {"pk": request.query_params.get("pk")}
        return request.build_absolute_uri(instance.content.get_url(section=instance, **kwargs))

    @staticmethod
    def get_name(instance):
        return instance.name if instance.name else instance.content.name



from django.contrib import admin

from appchance_sections.forms import SectionAdminForm


class SectionAdminMixin(admin.ModelAdmin):
    list_display = ("name", "content_name", "widget", "placement", "order", "content_url")
    form = SectionAdminForm



import django_filters
from django_filters.rest_framework import FilterSet


class SectionListFilter(FilterSet):
    placement = django_filters.CharFilter(method="get_by_placements", field_name="placement")

    @staticmethod
    def get_by_placements(queryset, _field_name, value):
        return queryset.filter(placement__in=value.split(","))

    class Meta:
        fields = ["placement"]



from django.forms import ChoiceField, ModelForm

from appchance_sections.utils import get_section_model

Section = get_section_model()


class SectionAdminForm(ModelForm):
    content_slug = ChoiceField(choices=[])
    widget = ChoiceField(choices=[])
    placement = ChoiceField(choices=[])

    class Meta:
        model = Section
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields.get("content_slug").choices = self._get_content_slug_choices()
        self.fields.get("widget").choices = self._get_widgets()
        self.fields.get("placement").choices = [(p, p) for p in Section.get_placements()]

    @staticmethod
    def _get_content_slug_choices():
        choices = set()
        for key, value in Section.get_contents().items():
            if value.prefix:
                label = f"[{value.prefix}] {value.name}"
            else:
                label = value.name
            choices.add((key, label))
        return sorted(list(choices), key=lambda x: x[1])

    @staticmethod
    def _get_widgets():
        choices = set()
        for content in Section.get_contents().values():
            for widget in content.widgets:
                if content.prefix:
                    label = f"[{content.prefix}] {widget}"
                else:
                    label = widget
                choices.add((widget, label))
        return sorted(list(choices), key=lambda x: x[1])



from typing import Callable, Dict, List, Set
from urllib.parse import unquote

from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from appchance_sections.cache import get_contents_cache, get_contents_cache_key, get_contents_cache_timeout


class Content:
    def __init__(
        self, slug, url, *, name=None, query_params=None, path_params=None, widgets=None, placements=None, prefix=None
    ):
        self.slug = slug
        self.name = name or slug
        self.url = url
        self.query_params = query_params or {}
        self.path_params = path_params or {}
        self.widgets = widgets or []
        self.placements = placements or []
        self.prefix = prefix

    def __str__(self):
        return self.slug

    def get_url(self, section=None, **kwargs):
        url = reverse(self.url, kwargs=self._get_path_params(**kwargs))
        query_string = urlencode(self._get_query_params(section))
        if query_string:
            return f"{url}?{query_string}"
        return url

    def _get_path_params(self, **kwargs):
        path_params = {}
        for key, val in self.path_params.items():
            if key in kwargs and kwargs[key] is not None:
                val = kwargs[key]
            elif not val:
                val = "{%s}" % key
            path_params[key] = val
        return path_params

    def _get_query_params(self, section):
        query_params = self.query_params
        if section and section.num_items > 0:
            query_params["page_size"] = section.num_items
        return query_params


class DynamicContentAbstract(models.Model):
    URL = ""
    QUERY_PARAMS: Dict[str, str] = {}
    WIDGETS: List[str] = []
    PLACEMENTS: List[str] = []
    PREFIX = ""
    NAME_FIELD = ""
    FILTER_ATTRIBUTE = ""

    name = models.CharField(_("Name"), max_length=255, blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.pk}"

    @property
    def content_slug(self):
        return f"{self.__class__.__name__}-{self.id}"

    @property
    def name_field(self):
        return getattr(self, self.NAME_FIELD) if self.NAME_FIELD and hasattr(self, self.NAME_FIELD) else str(self)

    @classmethod
    def get_cache_key(cls):
        return f"section_dynamic_content_{cls.__name__}"

    @classmethod
    def get_all_objects(cls):
        return cls.objects.all()


class SectionAbstract(models.Model):
    _contents: Dict[str, Content] = {}
    _content_callbacks: Set[Callable] = set()

    name = models.CharField(_("Name"), max_length=255, blank=True, null=True)

    content_slug = models.CharField(_("Content"), max_length=255)

    num_items = models.PositiveIntegerField(_("Number of items"), default=1)

    widget = models.CharField(_("Widget"), max_length=64, null=True, blank=True)
    placement = models.CharField(_("Placement"), max_length=64, null=True, blank=True)
    order = models.PositiveSmallIntegerField(_("Order"), default=0)

    is_active = models.BooleanField(_("Is active"), default=True)

    class Meta:
        abstract = True
        verbose_name = _("Section")
        verbose_name_plural = _("Sections")
        ordering = ["placement", "order"]

    def __str__(self):
        return f"id {self.id}"

    @classmethod
    def get_contents(cls) -> Dict[str, Content]:
        cache = get_contents_cache()
        key = get_contents_cache_key()
        timeout = get_contents_cache_timeout()
        contents = cache.get(key)
        if not contents:
            dynamic_contents = {}
            for callback in cls._content_callbacks:
                items = callback()
                for item in items:
                    dynamic_contents[item.slug] = item
            contents = {**cls._contents, **dynamic_contents}
            cache.set(key, contents, timeout)
        return contents

    @classmethod
    def get_widgets(cls):
        widgets: List[str] = []
        for content in cls.get_contents().values():
            widgets += content.widgets
        return list(set(widgets))

    @classmethod
    def get_placements(cls):
        placements: List[str] = []
        for content in cls.get_contents().values():
            placements += content.placements
        return list(set(placements))

    @property
    def content(self):
        return self.get_contents()[self.content_slug]

    def clean(self):
        error_dict = {}
        try:
            if self.widget not in self.content.widgets:
                error_dict["widget"] = ValidationError(
                    _("Can not set `%s` widget for `%s` content.") % (self.widget, self.content_name)
                )

            if self.placement not in self.content.placements:
                error_dict["placement"] = ValidationError(
                    _("Can not set `%s` placement for `%s` content.") % (self.placement, self.content_name)
                )

            if self.placement and self.widget:
                placement_prefix, *_ignore = self.placement.split("_")
                widget_prefix, *_ignore = self.widget.split("_")

                if placement_prefix != widget_prefix:
                    error_dict["widget"] = ValidationError(
                        _("Can not set `%s` widget for `%s` placement (`%s` is not equal `%s`).")
                        % (self.widget, self.placement, widget_prefix, placement_prefix)
                    )
        except KeyError:
            pass

        if error_dict:
            raise ValidationError(error_dict)

    @property
    def content_name(self):
        return self.content.name

    @property
    def content_url(self):
        return unquote(self.content.get_url(section=self))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SectionsConfig(AppConfig):
    name = "appchance_sections"
    label = "appchance_sections"
    verbose_name = _("Sections")



import logging

import django_filters
from rest_framework import mixins, viewsets
from rest_framework.permissions import AllowAny

from appchance_sections.filtersets import SectionListFilter
from appchance_sections.serializers import SectionSerializer
from appchance_sections.utils import get_section_model

logger = logging.getLogger(__name__)

Section = get_section_model()


class SectionViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    """List of all sections."""

    permission_classes = (AllowAny,)
    queryset = Section.objects.filter(is_active=True)
    serializer_class = SectionSerializer
    pagination_class = None
    filter_backends = [django_filters.rest_framework.DjangoFilterBackend]
    filter_class = SectionListFilter



from rest_framework import routers

from appchance_sections.views import SectionViewSet

router = routers.DefaultRouter()
router.register(r"sections", SectionViewSet)

app_name = "appchance_sections"  # pylint: disable=C0103
urlpatterns = router.urls
